import { map, merge, of, race, EMPTY } from 'rxjs';
import { defaultIfEmpty, finalize, mergeMap, shareReplay, tap, } from 'rxjs/operators';
import { bindProcClose } from './proc-close.js';
import { bindProcError } from './proc-error.js';
import { bindProcExit } from './proc-exit.js';
import { bindStderrData } from './stderr.js';
import { bindStdinData } from './stdin.js';
import { bindStdoutData } from './stdout.js';
export function bindEvent(proc, stderrMaxBufferSize, msgPrefixOpts, script, // for throw error
inputStream) {
    const { stderrPrefix } = msgPrefixOpts;
    const close$ = bindProcClose(proc);
    const exit$ = bindProcExit(proc);
    const closeOrExit$ = race(close$, exit$).pipe(map((data) => {
        const [exitCode, exitSignal] = data;
        if (typeof exitCode === 'number') {
            return { exitCode, exitSignal };
        }
        return { exitCode: 0, exitSignal };
    }), 
    // tap((codeSignal) => {
    //   console.log('close or exit event', codeSignal)
    // }),
    shareReplay(1));
    const stdout$ = bindStdoutData(proc.stdout, closeOrExit$);
    const error$ = bindProcError(proc, closeOrExit$);
    const skipUntilNotifier$ = closeOrExit$.pipe(mergeMap(({ exitCode }) => {
        return exitCode === 0 ? EMPTY : of(void 0);
    }));
    const stderr$ = bindStderrData(proc.stderr, closeOrExit$, skipUntilNotifier$, stderrMaxBufferSize).pipe(tap((row) => {
        const msg = row.data.toString();
        throw new Error(`${stderrPrefix} ${script}\n${msg}`);
    }));
    const stdin$ = inputStream && typeof inputStream.subscribe === 'function'
        ? bindStdinData(proc.stdin, inputStream)
        : EMPTY;
    const ret$ = merge(stdout$, stderr$, stdin$, error$, closeOrExit$).pipe(map((row) => {
        if (typeof row.exitCode === 'undefined') {
            return row;
        }
        const row2 = {
            ...row,
            data: Buffer.from(''),
        };
        return row2;
    }), finalize(() => {
        // sub process may not stop
        // link: https://nodejs.org/api/child_process.html#child_process_subprocess_kill_signal
        proc.killed || proc.kill();
    }), defaultIfEmpty({
        exitCode: 0,
        exitSignal: null,
        data: Buffer.from(''),
    }));
    return ret$;
}
//# sourceMappingURL=bindevent.js.map